# custom.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy

from .base import BaseSelection
from .base import BaseGraphSelection

class CustomSelection(BaseSelection):
	"""A selector based off a custom, user-defined feature-based function.

	This selector wraps a custom function that is passed in by the user. This
	function should take in a matrix containing a subset of the ground set and
	output the standard measure of quality of the subset. Each row in the matrix
	should be an example and each column should be a feature value for the 
	example (like a standard data matrix for machine learning).

	.. warning::
		If the function that the user wants to optimize is graph-based,
		i.e., that it determines the quality of a subset using similarities
		with other examples instead of the feature values directly, 
		e.g. facility location, CustomGraphSelection should be used instead. 
	
	.. note:: 
		Although apricot is built for submodular functions, there is no explicit
		restriction that the function passed in be submodular. Sometimes,
		supermodular functions can be reasonably maximized using the same
		greedy approaches used on submodular functions.


	Parameters
	----------
	n_samples : int
		The number of samples to return.

	func : function
		The feature-based set function to that this selector should wrap.

	initial_subset : list, numpy.ndarray or None
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'random' : randomly select elements (dummy optimizer)
			'modular' : approximate the function using its modular upper bound
			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'sample' : randomly take a subset and perform selection on that
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'two-stage'.

	optimizer_kwds : dict, optional
		Arguments to pass into the optimizer object upon initialization.
		Default is {}.

	function_kwds : dict, optional
		Arguments to pass into the function object for each call.
		Default is {}.

	n_jobs : int, optional
		The number of cores to use for processing. This value is multiplied
		by 2 when used to set the number of threads. If set to -1, use all
		cores and threads. Default is -1.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, function, initial_subset=None, 
		optimizer='two-stage', optimizer_kwds={}, function_kwds={},
		n_jobs=1, random_state=None, verbose=False):

		if callable(function) == False:
			raise ValueError("Passed in function must be callable.")

		self.function = function
		self.function_kwds = function_kwds

		super(CustomSelection, self).__init__(n_samples=n_samples, 
			initial_subset=initial_subset, optimizer=optimizer, 
			optimizer_kwds=optimizer_kwds, n_jobs=n_jobs, 
			random_state=random_state, verbose=verbose) 

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select the examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : CustomSelection
			The fit step returns this selector object.
		"""

		return super(CustomSelection, self).fit(X, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

	def _initialize(self, X):
		super(CustomSelection, self)._initialize(X)

		if self.initial_subset is None:
			pass
		elif self.initial_subset.ndim == 2:
			if self.initial_subset.shape[1] != X.shape[1]:
				raise ValueError("The number of columns in the initial subset must " \
					"match the number of columns in X.")
		elif self.initial_subset.ndim == 1:
			self.initial_subset = X[self.initial_subset]
		else:
			raise ValueError("The initial subset must be either a two dimensional" \
				" matrix of examples or a one dimensional mask.")

		if self.initial_subset is None:
			self.total_gain = 0
		else:
			self.total_gain = self.func(self.initial_subset)


	def _calculate_gains(self, X, idxs=None):
		"""This function will return the gain that each example would give.

		This function will return the gains that each example would give if
		added to the selected set. When a matrix of examples is given, a
		vector will be returned showing the gain for each example. When
		a single element is passed in, it will return a singe value."""

		idxs = idxs if idxs is not None else self.idxs
		gains = numpy.zeros(idxs.shape[0], dtype='float64')

		x0 = numpy.zeros((1, X.shape[1]))

		if self.initial_subset is not None:
			X_ = numpy.concatenate([self.initial_subset, self.subset, x0])
		else:
			X_ = numpy.concatenate([self.subset, x0])

		for i, idx in enumerate(idxs):
			X_[-1] = X[idx]
			gains[i] = self.function(X_, **self.function_kwds) - self.total_gain

		return gains

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		"""This function will update the internal statistics from a stream.

		This function will update the various internal statistics that are a
		part of the sieve algorithm for streaming submodular optimization. This
		function does not directly return gains but it updates the values
		used by a streaming optimizer.
		"""

		super(CustomSelection, self)._calculate_sieve_gains(X,
			thresholds, idxs)

		raise NotImplementedError

	def _select_next(self, X, gain, idx):
		"""This function will add the given item to the selected set."""

		self.total_gain += gain

		super(CustomSelection, self)._select_next(
			X, gain, idx)


class CustomGraphSelection(BaseGraphSelection):
	"""A selector based off a custom, user-defined graph-based function.

	This selector wraps a custom graph-based function that is passed in by the user. 
	This function should take in a matrix containing a subset of the ground set and
	output the standard measure of quality of the subset. Each row in the matrix
	should be an example and each column should be the similarity to each element
	in the ground set, with the number of columns being equal to the size of the
	ground set.
	
	.. note:: 
		Although apricot is built for submodular functions, there is no explicit
		restriction that the function passed in be submodular. Sometimes,
		supermodular functions can be reasonably maximized using the same
		greedy approaches used on submodular functions.


	Parameters
	----------
	n_samples : int
		The number of samples to return.

	func : function
		The graph-based set function to that this selector should wrap.

	initial_subset : list, numpy.ndarray or None
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'random' : randomly select elements (dummy optimizer)
			'modular' : approximate the function using its modular upper bound
			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'sample' : randomly take a subset and perform selection on that
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'two-stage'.

	optimizer_kwds : dict, optional
		Arguments to pass into the optimizer object upon initialization.
		Default is {}.

	function_kwds : dict, optional
		Arguments to pass into the function object for each call.
		Default is {}.

	n_jobs : int, optional
		The number of cores to use for processing. This value is multiplied
		by 2 when used to set the number of threads. If set to -1, use all
		cores and threads. Default is -1.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, function, metric='euclidean', 
		initial_subset=None, optimizer='two-stage', optimizer_kwds={}, 
		function_kwds={}, n_jobs=1, random_state=None, verbose=False):

		if callable(function) == False:
			raise ValueError("Passed in function must be callable.")

		self.function = function
		self.function_kwds = function_kwds

		super(CustomGraphSelection, self).__init__(n_samples=n_samples, 
			metric=metric, initial_subset=initial_subset, optimizer=optimizer, 
			optimizer_kwds=optimizer_kwds, n_jobs=n_jobs, 
			random_state=random_state, verbose=verbose) 

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select the examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : CustomSelection
			The fit step returns this selector object.
		"""

		return super(CustomGraphSelection, self).fit(X, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

	def _initialize(self, X):
		super(CustomGraphSelection, self)._initialize(X)

		if self.initial_subset is None:
			pass
		elif self.initial_subset.ndim == 2:
			if self.initial_subset.shape[1] != X.shape[1]:
				raise ValueError("The number of columns in the initial subset must " \
					"match the number of columns in X.")
		elif self.initial_subset.ndim == 1:
			self.initial_subset = X[self.initial_subset]
		else:
			raise ValueError("The initial subset must be either a two dimensional" \
				" matrix of examples or a one dimensional mask.")

		if self.initial_subset is None:
			self.total_gain = 0
		else:
			self.total_gain = self.func(self.initial_subset)


	def _calculate_gains(self, X, idxs=None):
		"""This function will return the gain that each example would give.

		This function will return the gains that each example would give if
		added to the selected set. When a matrix of examples is given, a
		vector will be returned showing the gain for each example. When
		a single element is passed in, it will return a singe value."""

		idxs = idxs if idxs is not None else self.idxs
		gains = numpy.zeros(idxs.shape[0], dtype='float64')

		x0 = numpy.zeros((1, X.shape[1]))

		if self.initial_subset is not None:
			X_ = numpy.concatenate([self.initial_subset, self.subset, x0])
		else:
			X_ = numpy.concatenate([self.subset, x0])

		for i, idx in enumerate(idxs):
			X_[-1] = X[idx]
			gains[i] = self.function(X_, **self.function_kwds) - self.total_gain

		return gains

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		"""This function will update the internal statistics from a stream.

		This function will update the various internal statistics that are a
		part of the sieve algorithm for streaming submodular optimization. This
		function does not directly return gains but it updates the values
		used by a streaming optimizer.
		"""

		super(CustomGraphSelection, self)._calculate_sieve_gains(X,
			thresholds, idxs)

		raise NotImplementedError

	def _select_next(self, X, gain, idx):
		"""This function will add the given item to the selected set."""

		self.total_gain += gain

		super(CustomGraphSelection, self)._select_next(
			X, gain, idx)



# mixture.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>
	
import numpy

from .base import BaseSelection
from .base import BaseGraphSelection
from ..utils import _calculate_pairwise_distances

from tqdm import tqdm

class MixtureSelection(BaseSelection):
	"""A selection approach based on a mixture of submodular functions.

	A convenient property of submodular functions is that any linear 
	combination of submodular functions is still submodular. More generally, 
	the linear combination of any number of submodular functions (assuming 
	non-negative weights) is still a submodular function. Because of this, 
	a mixture of submodular functions can be optimized using the same 
	algorithms as an individual submodular function. Mixtures can be 
	useful in situations where there are different important aspects 
	of data that are each submodular.

	The general form of a mixture function is 

	.. math::
		f(X) = \\sum\\limits_{i=1}^{M} \\alpha_{i} g_{i}(X) 

	where :math:`f` indicates the mixture function, :math:`M` is the number of 
	functions in the mixture, :math:`X` is a subset, :math:`\\alpha_{i}` is the 
	weight of the :math:`i`-th function and :math:`g_{i}` is the :math:`i`-th 
	function.

	.. note::
		There must be at least two components to the mixture, and all 
		:math:`\\alpha` must be non-negative.

	This class can also be used to add regularizers to the selection procedure.
	If a submodular function is mixed with another submodular function that
	acts as a regularizer, such as feature based selection mixed with a
	custom function measuring some property of the selected subset.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	functions : list
		The list of submodular functions to mix together. The submodular
		functions should be instantiated.

	weights : list, numpy.ndarray or None, optional
		The relative weight of each submodular function. This is the value
		that the gain from each submodular function is multiplied by before
		being added together. The default is equal weight for each function.

	initial_subset : list, numpy.ndarray or None, optional
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'random' : randomly select elements (dummy optimizer)
			'modular' : approximate the function using its modular upper bound
			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'sample' : randomly take a subset and perform selection on that
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'two-stage'.

	optimizer_kwds : dict or None
		A dictionary of arguments to pass into the optimizer object. The keys
		of this dictionary should be the names of the parameters in the optimizer
		and the values in the dictionary should be the values that these
		parameters take. Default is None.

	reservoir : numpy.ndarray or None
		The reservoir to use when calculating gains in the sieve greedy
		streaming optimization algorithm in the `partial_fit` method.
		Currently only used for graph-based functions. If a numpy array
		is passed in, it will be used as the reservoir. If None is passed in,
		will use reservoir sampling to collect a reservoir. Default is None.

	max_reservoir_size : int 
		The maximum size that the reservoir can take. If a reservoir is passed
		in, this value is set to the size of that array. Default is 1000.

	n_jobs : int
		The number of threads to use when performing computation in parallel.
		Currently, this parameter is exposed but does not actually do anything.
		This will be fixed soon.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool, optional
		Whether to print output during the selection process.

	Attributes
	----------
	pq : PriorityQueue
		The priority queue used to implement the lazy greedy algorithm.

	n_samples : int
		The number of samples to select.

	submodular_functions : list
		A concave function for transforming feature values, often referred to as
		phi in the literature.

	weights : numpy.ndarray
		The weights of each submodular function.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, functions, weights=None, metric='ignore',
		initial_subset=None, optimizer='two-stage', optimizer_kwds={}, n_neighbors=None, 
		reservoir=None, max_reservoir_size=1000, n_jobs=1, random_state=None, 
		verbose=False):

		if len(functions) < 2:
			raise ValueError("Must mix at least two functions.")

		self.m = len(functions)
		self.functions = functions

		if weights is None:
			self.weights = numpy.ones(self.m, dtype='float64')
		else:
			self.weights = numpy.array(weights, dtype='float64', copy=False)

		super(MixtureSelection, self).__init__(n_samples=n_samples, 
			initial_subset=initial_subset, optimizer=optimizer, 
			optimizer_kwds=optimizer_kwds, reservoir=reservoir,
			max_reservoir_size=max_reservoir_size, n_jobs=n_jobs, 
			random_state=random_state, verbose=verbose)

		self.metric = metric.replace("corr", "correlation")
		self.n_neighbors = n_neighbors

		for function in self.functions:
			function.initial_subset = self.initial_subset
			function.reservoir = reservoir
			function.max_reservoir_size = max_reservoir_size
			function.metric = 'precomputed'

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select the examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : MixtureSelection
			The fit step returns this selector object.
		"""

		# If self.metric is ignore, this will return the same matrix.
		# Otherwise, it will convert it to a pairwise similarity matrix.
		self._X = X
		X = _calculate_pairwise_distances(X, metric=self.metric, 
			n_neighbors=self.n_neighbors)

		return super(MixtureSelection, self).fit(X, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

	def _initialize(self, X):
		super(MixtureSelection, self)._initialize(X)

		for function in self.functions:
			function._initialize(X)

	def _calculate_gains(self, X, idxs=None):
		"""This function will return the gain that each example would give.

		This function will return the gains that each example would give if
		added to the selected set. When a matrix of examples is given, a
		vector will be returned showing the gain for each example. When
		a single element is passed in, it will return a singe value."""

		idxs = idxs if idxs is not None else self.idxs
		gains = numpy.zeros(idxs.shape[0], dtype='float64')

		for i, function in enumerate(self.functions):
			gains += function._calculate_gains(X, idxs) * self.weights[i]

		return gains

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		super(MixtureSelection, self)._calculate_sieve_gains(X, 
			thresholds, idxs)

		for function in self.functions:
			super(function.__class__, function)._calculate_sieve_gains(X,
				thresholds, idxs)

		t = numpy.zeros(thresholds.shape[0], dtype='float64') - 1

		for j in range(X.shape[0]):
			x = X[j:j+1]

			current_values = []
			total_gains = []

			gain = numpy.zeros((len(thresholds), self.n_samples), dtype='float64')
			for i, function in enumerate(self.functions):
				current_values.append(function.sieve_current_values_.copy())
				total_gains.append(function.sieve_total_gains_.copy())

				function._calculate_sieve_gains(x, t, idxs)
				gain += function.sieve_gains_ * self.weights[i]

			for l in range(len(thresholds)):
				if self.sieve_n_selected_[l] == self.n_samples:
					continue

				threshold = ((thresholds[l] / 2. - self.sieve_total_gains_[l]) 
					/ (self.n_samples - self.sieve_n_selected_[l]))

				if gain[l, self.sieve_n_selected_[l]] > threshold:
					self.sieve_total_gains_[l] += gain[l, self.sieve_n_selected_[l]]
					self.sieve_selections_[l, self.sieve_n_selected_[l]] = idxs[j]
					self.sieve_gains_[l, self.sieve_n_selected_[l]] = gain[l, self.sieve_n_selected_[l]]
					self.sieve_n_selected_[l] += 1
				else:
					v = self.sieve_n_selected_[l]

					for i, function in enumerate(self.functions):
						vi = function.sieve_n_selected_[l]
						if vi != self.sieve_n_selected_[l]:
							function.sieve_current_values_[l] = current_values[i][l]
							function.sieve_total_gains_[l] = total_gains[i][l]
							function.sieve_n_selected_[l] = vi - 1
							if vi < self.n_samples:
								function.sieve_selections_[l, vi] = 0
								function.sieve_gains_[l, vi] = 0
								function.sieve_subsets_[l, vi] = 0

	def _select_next(self, X, gain, idx):
		"""This function will add the given item to the selected set."""

		for function in self.functions:
			function._select_next(X, gain, idx)

		super(MixtureSelection, self)._select_next(X, gain, idx)



# __init__.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

from .functions import FeatureBasedSelection
from .functions import MaxCoverageSelection

from .functions import FacilityLocationSelection
from .functions import SaturatedCoverageSelection
from .functions import SumRedundancySelection
from .functions import GraphCutSelection

from .functions import MixtureSelection

from .functions import BaseSelection
from .functions import BaseGraphSelection

from .functions import CustomSelection
from .functions import CustomGraphSelection


__version__ = '0.6.0'



# base.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com> 

"""
This file contains code that implements the core of the submodular selection
algorithms.
"""

import numpy
from tqdm import tqdm

from ..optimizers import BaseOptimizer
from ..optimizers import NaiveGreedy
from ..optimizers import LazyGreedy
from ..optimizers import ApproximateLazyGreedy
from ..optimizers import TwoStageGreedy
from ..optimizers import StochasticGreedy
from ..optimizers import BidirectionalGreedy
from ..optimizers import GreeDi
from ..optimizers import SieveGreedy
from ..optimizers import OPTIMIZERS

from ..utils import PriorityQueue
from ..utils import check_random_state
from ..utils import _calculate_pairwise_distances

from scipy.sparse import csr_matrix


class BaseSelection(object):
	"""The base selection object.

	This object defines the structures that all submodular selection algorithms
	should follow. All algorithms will have the same public methods and the
	same attributes.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	initial_subset : list, numpy.ndarray or None, optional
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'naive'.

	optimizer_kwds : dict or None
		A dictionary of arguments to pass into the optimizer object. The keys
		of this dictionary should be the names of the parameters in the optimizer
		and the values in the dictionary should be the values that these
		parameters take. Default is None.

	reservoir : numpy.ndarray or None
		The reservoir to use when calculating gains in the sieve greedy
		streaming optimization algorithm in the `partial_fit` method.
		Currently only used for graph-based functions. If a numpy array
		is passed in, it will be used as the reservoir. If None is passed in,
		will use reservoir sampling to collect a reservoir. Default is None.

	max_reservoir_size : int 
		The maximum size that the reservoir can take. If a reservoir is passed
		in, this value is set to the size of that array. Default is 1000.

	n_jobs : int
		The number of threads to use when performing computation in parallel.
		Currently, this parameter is exposed but does not actually do anything.
		This will be fixed soon.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, initial_subset=None, optimizer='lazy', 
		optimizer_kwds={}, reservoir=None, max_reservoir_size=1000, 
		n_jobs=1, random_state=None, verbose=False):
		if n_samples <= 0:
			raise ValueError("n_samples must be a positive value.")

		if not isinstance(initial_subset, (list, numpy.ndarray)) and initial_subset is not None: 
			raise ValueError("initial_subset must be a list, numpy array, or None")
		if isinstance(initial_subset, (list, numpy.ndarray)):
			initial_subset = numpy.array(initial_subset)

		if not isinstance(optimizer, BaseOptimizer):
			if optimizer not in OPTIMIZERS.keys():
				raise ValueError("Optimizer must be an optimizer object or " \
					"a str in {}.".format(str(OPTIMIZERS.keys())))

		if isinstance(optimizer, BaseOptimizer):
			optimizer.function = self

		if verbose not in (True, False):
			raise ValueError("verbosity must be True or False")

		self.n_samples = n_samples
		self.metric = 'ignore'
		self.random_state = check_random_state(random_state)
		self.optimizer = optimizer
		self.optimizer_kwds = optimizer_kwds
		self.n_jobs = n_jobs
		self.verbose = verbose
		self.initial_subset = initial_subset

		self.ranking = None
		self.idxs = None
		self.gains = None
		self.subset = None
		self.sparse = None
		self._X = None
		
		self.sieve_current_values_ = None
		self.n_seen_ = 0
		self.reservoir_size = 0 if reservoir is None else reservoir.shape[0]
		self.reservoir = reservoir
		self.max_reservoir_size = max_reservoir_size if reservoir is None else reservoir.shape[0]
		self.update_reservoir_ = reservoir is None

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select a subset of examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : BaseGraphSelection
			The fit step returns this selector object.
		"""

		allowed_dtypes = list, numpy.ndarray, csr_matrix

		if not isinstance(X, allowed_dtypes):
			raise ValueError("X must be either a list of lists, a 2D numpy " \
				"array, or a scipy.sparse.csr_matrix.")
		if isinstance(X, numpy.ndarray) and len(X.shape) != 2:
			raise ValueError("X must have exactly two dimensions.")
		if numpy.min(X) < 0.0 and numpy.max(X) > 0.:
			raise ValueError("X cannot contain negative values or must be entirely "\
				"negative values.")
		if self.n_samples > X.shape[0]:
			raise ValueError("Cannot select more examples than the number in" \
				" the data set.")

		if not self.sparse:
			if X.dtype != 'float64':
				X = X.astype('float64')

		if isinstance(self.optimizer, str):
			optimizer = OPTIMIZERS[self.optimizer](function=self, 
				verbose=self.verbose, random_state=self.random_state,
				**self.optimizer_kwds)
		else:
			optimizer = self.optimizer

		self._X = X if self._X is None else self._X
		self._initialize(X)

		if self.verbose:
			self.pbar = tqdm(total=self.n_samples, unit_scale=True)

		optimizer.select(X, self.n_samples, sample_cost=sample_cost)

		if self.verbose == True:
			self.pbar.close()

		self.ranking = numpy.array(self.ranking)
		self.gains = numpy.array(self.gains)
		return self

	def partial_fit(self, X, y=None, sample_weight=None, sample_cost=None):
		allowed_dtypes = list, numpy.ndarray, csr_matrix

		if not isinstance(X, allowed_dtypes):
			raise ValueError("X must be either a list of lists, a 2D numpy " \
				"array, or a scipy.sparse.csr_matrix.")
		if isinstance(X, numpy.ndarray) and len(X.shape) != 2:
			raise ValueError("X must have exactly two dimensions.")

		if not self.sparse:
			if X.dtype != 'float64':
				X = X.astype('float64')

		if not isinstance(self.optimizer, SieveGreedy):
			self.optimizer = OPTIMIZERS['sieve'](function=self, 
				verbose=self.verbose, random_state=self.random_state,
				**self.optimizer_kwds)

		self._X = X if self._X is None else self._X
		self._initialize(X)

		if self.verbose:
			self.pbar = tqdm(total=self.n_samples, unit_scale=True)

		self.optimizer.select(X, self.n_samples, sample_cost=sample_cost)

		if self.verbose == True:
			self.pbar.close()

		self.ranking = numpy.array(self.ranking)
		self.gains = numpy.array(self.gains)
		self._X = None
		return self

	def transform(self, X, y=None, sample_weight=None):
		"""Transform a data set to include only the selected examples.

		This method will return a selection of X and optionally selections
		of y and sample_weight. The default setting is to select items based
		on the ranking determined in the `fit` step with examples in the same
		order as that ranking. Optionally, the whole data set can be returned,
		with the weights corresponding to samples that were not selected set
		to 0. This setting can be controlled by setting `pipeline=True`. 

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and the corresponding labels for the rows that have
			been selected. Default is None. 

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The sample weights to transform. If passed in this function will
			return the selected labels (y) and the selected samples, even
			if no labels were passed in. Default is None.

		Returns
		-------
		X_subset : numpy.ndarray, shape=(n_samples, d)
			A subset of the data such that n_samples < n and n_samples is the
			integer provided at initialization.

		y_subset : numpy.ndarray, shape=(n_samples,), optional
			The labels that match with the indices of the samples if y is
			passed in. Only returned if passed in.

		sample_weight_subset : numpy.ndarray, shape=(n_samples,), optional
			The weight of each example.
		"""

		r = self.ranking

		if sample_weight is not None:
			if y is None:
				return X[r], None, sample_weight[r]
			else:
				return X[r], y[r], sample_weight[r]

		else:
			if y is None:
				return X[r]
			else:
				return X[r], y[r]

	def fit_transform(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run optimization and select a subset of examples.

		This method will first perform the `fit` step and then perform the
		`transform` step, returning a transformed data set. 

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and the corresponding labels for the rows that have
			been selected. Default is None. 

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The sample weights to transform. If passed in this function will
			return the selected labels (y) and the selected samples, even
			if no labels were passed in. Default is None.

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		X_subset : numpy.ndarray, shape=(n_samples, d)
			A subset of the data such that n_samples < n and n_samples is the
			integer provided at initialization.

		y_subset : numpy.ndarray, shape=(n_samples,), optional
			The labels that match with the indices of the samples if y is
			passed in. Only returned if passed in.

		sample_weight_subset : numpy.ndarray, shape=(n_samples,), optional
			The weight of each example.
		"""

		return self.fit(X, y=y, sample_weight=sample_weight, 
			sample_cost=sample_cost).transform(X, y=y, 
			sample_weight=sample_weight)

	def _initialize(self, X, idxs=None):
		n, d = X.shape
		self._X = X if self._X is None else self._X

		self.sparse = isinstance(X, csr_matrix)
		self.ranking = []
		self.gains = []
		self.subset = numpy.zeros((0, self._X.shape[1]), dtype='float64')

		self.current_values = numpy.zeros(d, dtype='float64')
		self.current_concave_values = numpy.zeros(d, dtype='float64')
		self.mask = numpy.zeros(n, dtype='int8')

		if self.initial_subset is not None:
			if self.initial_subset.ndim == 1:
				if self.initial_subset.dtype == bool:
					self.initial_subset = numpy.where(self.initial_subset == 1)[0]
				
				if len(self.initial_subset) + self.n_samples > X.shape[0]:
					raise ValueError("When using a mask for the initial subset" \
						" must selected fewer than the size of the subset minus" \
						" the initial subset size, i.e., n_samples < X.shape[0] -"\
						" initial_subset.shape[0].")

				if self.initial_subset.max() > X.shape[0]:
					raise ValueError("When passing in an integer mask for the initial subset"\
						" the maximum value cannot exceed the size of the data set.")
				elif self.initial_subset.min() < 0:
					raise ValueError("When passing in an integer mask for the initial subset"\
						" the minimum value cannot be negative.")
				
				self.mask[self.initial_subset] = 1

		self.idxs = numpy.where(self.mask == 0)[0]

	def _calculate_gains(self, X, idxs=None):
		raise NotImplementedError

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		n = X.shape[0]
		d = X.shape[1] if self.reservoir is None else self.max_reservoir_size
		l = len(thresholds)

		if self.sieve_current_values_ is None:
			self.sieve_current_values_ = numpy.zeros((l, d), 
				dtype='float64')
			self.sieve_selections_ = numpy.zeros((l, self.n_samples), 
				dtype='int64') - 1
			self.sieve_gains_ = numpy.zeros((l, self.n_samples), 
				dtype='float64') - 1
			self.sieve_n_selected_ = numpy.zeros(l, 
				dtype='int64')
			self.sieve_total_gains_ = numpy.zeros(l, 
				dtype='float64')
			self.sieve_subsets_ = numpy.zeros((l, self.n_samples, 
				self._X.shape[1]), dtype='float64')
		else:
			j = l - self.sieve_current_values_.shape[0]
			if j > 0:
				self.sieve_current_values_ = numpy.vstack([
					self.sieve_current_values_, numpy.zeros((j, d), 
						dtype='float64')])
				self.sieve_selections_ = numpy.vstack([
					self.sieve_selections_, numpy.zeros((j, self.n_samples), 
						dtype='int64') - 1])
				self.sieve_gains_ = numpy.vstack([self.sieve_gains_, 
					numpy.zeros((j, self.n_samples), dtype='float64')])
				self.sieve_n_selected_ = numpy.concatenate([
					self.sieve_n_selected_, numpy.zeros(j, dtype='int64')])
				self.sieve_total_gains_ = numpy.concatenate([
					self.sieve_total_gains_, numpy.zeros(j, dtype='float64')])
				self.sieve_subsets_ = numpy.concatenate([self.sieve_subsets_, 
					numpy.zeros((j, self.n_samples, self._X.shape[1]), 
						dtype='float64')])

	def _select_next(self, X, gain, idx):
		self.ranking.append(idx)
		self.gains.append(gain)
		self.mask[idx] = True
		self.idxs = numpy.where(self.mask == 0)[0]

		if self.sparse:
			X = self._X[idx:idx+1].toarray()
		else:
			X = self._X[idx:idx+1]

		if self.metric != 'precomputed':
			self.subset = numpy.concatenate([self.subset, X])


class BaseGraphSelection(BaseSelection):
	"""The base graph selection object.

	This object defines the structures that all submodular selection algorithms
	should follow if they operate on a graph, such as pairwise similarity 
	measurements. All algorithms will have the same public methods and the same 
	attributes.

	NOTE: All ~pairwise~ values in your data must be positive for these
	selection methods to work.

	This implementation allows users to pass in either their own symmetric
	square matrix of similarity values, or a data matrix as normal and a function
	that calculates these pairwise values.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	metric : str
		The method for converting a data matrix into a square symmetric matrix
		of pairwise similarities. If a string, can be any of the metrics
		implemented in sklearn (see https://scikit-learn.org/stable/modules/
		generated/sklearn.metrics.pairwise_distances.html), including
		"precomputed" if one has already generated a similarity matrix. Note
		that sklearn calculates distance matrices whereas apricot operates on
		similarity matrices, and so a distances.max() - distances transformation
		is performed on the resulting distances. For backcompatibility,
		'corr' will be read as 'correlation'.

	initial_subset : list, numpy.ndarray or None
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'naive'.

	optimizer_kwds : dict or None
		A dictionary of arguments to pass into the optimizer object. The keys
		of this dictionary should be the names of the parameters in the optimizer
		and the values in the dictionary should be the values that these
		parameters take. Default is None.

	n_neighbors : int or None
		When constructing a similarity matrix, the number of nearest neighbors
		whose similarity values will be kept. The result is a sparse similarity
		matrix which can significantly speed up computation at the cost of
		accuracy. Default is None.

	reservoir : numpy.ndarray or None
		The reservoir to use when calculating gains in the sieve greedy
		streaming optimization algorithm in the `partial_fit` method.
		Currently only used for graph-based functions. If a numpy array
		is passed in, it will be used as the reservoir. If None is passed in,
		will use reservoir sampling to collect a reservoir. Default is None.

	max_reservoir_size : int 
		The maximum size that the reservoir can take. If a reservoir is passed
		in, this value is set to the size of that array. Default is 1000.

	n_jobs : int
		The number of threads to use when performing computation in parallel.
		Currently, this parameter is exposed but does not actually do anything.
		This will be fixed soon.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	metric : callable
		A function that takes in a data matrix and converts it to a square
		symmetric matrix.

	ranking : numpy.array int
		The selected samples in the order of their gain.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, metric='euclidean', 
		initial_subset=None, optimizer='two-stage', optimizer_kwds={},
		n_neighbors=None, reservoir=None, max_reservoir_size=1000, 
		n_jobs=1, random_state=None, verbose=False):

		super(BaseGraphSelection, self).__init__(n_samples=n_samples, 
			initial_subset=initial_subset, optimizer=optimizer, 
			optimizer_kwds=optimizer_kwds, reservoir=reservoir, 
			max_reservoir_size=max_reservoir_size, n_jobs=n_jobs, 
			random_state=random_state, verbose=verbose)

		self.metric = metric.replace("corr", "correlation")
		self.n_neighbors = n_neighbors


	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select a subset of examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : BaseGraphSelection
			The fit step returns this selector object.
		"""

		if isinstance(X, csr_matrix) and self.metric not in ("precomputed", "ignore"):
			raise ValueError("Must passed in a precomputed sparse " \
				"similarity matrix or a dense feature matrix.")
		if self.metric == 'precomputed' and X.shape[0] != X.shape[1]:
			raise ValueError("Precomputed similarity matrices " \
				"must be square and symmetric.")

		X_pairwise = _calculate_pairwise_distances(X, metric=self.metric, 
			n_neighbors=self.n_neighbors)
	
		self._X = X
		return super(BaseGraphSelection, self).fit(X_pairwise, y=y,
			sample_weight=sample_weight, sample_cost=sample_cost)

	def partial_fit(self, X, y=None, sample_weight=None, sample_cost=None):
		if self.reservoir is None:
			self.reservoir = numpy.empty((self.max_reservoir_size, X.shape[1]))

		if self.update_reservoir_:
			for i in range(X.shape[0]):
				if self.reservoir_size < self.max_reservoir_size:
					self.reservoir[self.reservoir_size] = X[i]
					self.reservoir_size += 1
				else:
					r = self.random_state.choice(self.n_seen_ + i)
					if r < self.max_reservoir_size:
						self.reservoir[r] = X[i]
						#self.current_values_[:, r] = 0.

		X_pairwise = _calculate_pairwise_distances(X, 
			Y=self.reservoir[:self.reservoir_size], metric=self.metric)

		self._X = X
		super(BaseGraphSelection, self).partial_fit(X_pairwise, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

		self.current_values = numpy.zeros(self.reservoir_size, 
			dtype='float64')
		self.n_seen_ += X.shape[0]

	def _initialize(self, X_pairwise, idxs=None):
		super(BaseGraphSelection, self)._initialize(X_pairwise, idxs=idxs)

	def _calculate_gains(self, X_pairwise):
		super(BaseGraphSelection, self)._calculate_gains(X_pairwise)

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		super(BaseGraphSelection, self)._calculate_sieve_gains(X, thresholds,
			idxs)

	def _select_next(self, X_pairwise, gain, idx):
		super(BaseGraphSelection, self)._select_next(X_pairwise, gain, idx)



# __init__.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

from .featureBased import FeatureBasedSelection
from .maxCoverage import MaxCoverageSelection

from .facilityLocation import FacilityLocationSelection
from .saturatedCoverage import SaturatedCoverageSelection
from .sumRedundancy import SumRedundancySelection
from .graphCut import GraphCutSelection

from .mixture import MixtureSelection

from .custom import CustomSelection
from .custom import CustomGraphSelection

from .base import BaseSelection
from .base import BaseGraphSelection


# optimizers.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import os
import copy
import numpy
import scipy

from tqdm import tqdm

from .utils import PriorityQueue
from .utils import check_random_state
from .utils import _calculate_pairwise_distances

class BaseOptimizer(object):
	"""An approach for optimizing submodular functions.

	This object contains a single method, `select`, that implements an
	algorithm for optimizing submodular functions. This method will take in the
	data set, the number of examples to select, and a stored reference to the
	submodular function to be optimized, and yield a subset of examples. The
	method operates on the gains provided in the `_calculate_gains` method from
	the submodular function object and stores examples using the `_select_next`
	method.

	Parameters
	----------
	function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection or None
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized. If None, will be set by the selector when passed in.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.
	"""

	def __init__(self, function=None, random_state=None, n_jobs=None, 
		verbose=False):
		self.function = function
		self.random_state = check_random_state(random_state)
		self.n_jobs = n_jobs
		self.verbose = verbose
		self.gains_ = None

	def select(self, X, k):
		raise NotImplementedError


class NaiveGreedy(BaseOptimizer):
	"""The naive greedy algorithm for optimization.

	The naive greedy algorithm is the simplest greedy approach for optimizing 
	submodular functions. The approach simply iterates through each example in 
	the ground set that has not already been selected and calculates the gain 
	in function value that would result from adding that example to the 
	selected set. This process is embarassingly parallel and so is extremely 
	amenable both to parallel processing and distributed computing. Further, 
	because it is conceptually simple, it is also simple to implement.

	The naive greedy algorithm can be specified for any function by passing in 
	`optimizer='naive'` to the relevant selector object. Here is an example of 
	specifying the naive greedy algorithm for optimizing a feature-based function.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='naive')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, random_state=None, n_jobs=None, 
		verbose=False):
		super(NaiveGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		while cost < k:
			gains = self.function._calculate_gains(X) / sample_cost[self.function.idxs]
			idxs = numpy.lexsort((numpy.arange(gains.shape[0]), -gains))

			for idx in idxs:
				best_idx = self.function.idxs[idx]
				if cost + sample_cost[best_idx] <= k:
					break
			else:
				break

			cost += sample_cost[best_idx]
			gain = gains[idx] * sample_cost[best_idx]
			self.function._select_next(X[best_idx], gain, best_idx)

			if self.verbose == True:
				self.function.pbar.update(round(sample_cost[best_idx], 1))


class LazyGreedy(BaseOptimizer):
	"""The lazy/accelerated greedy algorithm for optimization.

	The lazy (or accelerated) greedy algorithm is a fast alternate to the 
	naive greedy algorithm that results in the same items being selected. 
	This algorithm exploits the diminishing returns property of submodular 
	functions in order to avoid re-evaluating examples that are known to 
	provide little gain. If an example has a small gain relative to other 
	examples, it is unlikely to be the next selected example because that 
	gain can only go down as more items are selected. Thus, the example 
	should only be re-evaluated once the gains of other examples have gotten 
	smaller.

	The key idea of the lazy greedy algorithm is to maintain a priority queue 
	where the examples are the elements in the queue and the priorities are 
	the gains the last time they were evaluated. The algorithm has two steps. 
	The first step is to calculate the gain that each example would have if 
	selected first (the modular upper bound) and populate the priority queue 
	using these values. The second step is to recalculate the gain of the 
	first example in the priority queue and then add the example back into 
	the queue. If the example remains at the front of the queue it is selected 
	because no other example could have a larger gain once re-evaluated (due 
	to the diminishing returns property).

	While the worst case time complexity of this algorithm is the same as the 
	naive greedy algorithm, in practice it can be orders of magnitude faster. 
	Empirically, it appears to accelerate graph-based functions much more 
	than it does feature-based ones. Functions also seem to be more 
	accelerated the more curved they are.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='lazy')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.pq : utils.PriorityQueue
		The priority queue used to order examples for evaluation.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, random_state=None, n_jobs=None, 
		verbose=False):
		super(LazyGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		gains = self.function._calculate_gains(X) / sample_cost[self.function.idxs]
		self.pq = PriorityQueue(self.function.idxs, -gains)

		while cost < k:
			best_gain = float("-inf")
			best_idx = None
			
			while True:
				if len(self.pq.pq) == 0:
					return

				prev_gain, idx = self.pq.pop()
				prev_gain = -prev_gain

				if cost + sample_cost[idx] > k:
					continue

				if best_idx == idx:
					break
				
				idxs = numpy.array([idx])
				gain = self.function._calculate_gains(X, idxs)[0] / sample_cost[idx]
				self.pq.add(idx, -gain)

				if gain > best_gain:
					best_gain = gain
					best_idx = idx
				elif gain == best_gain and best_gain == 0.0:
					best_gain = gain
					best_idx = idx
					break

			cost += sample_cost[best_idx]
			best_gain *= sample_cost[best_idx]
			self.function._select_next(X[best_idx], best_gain, best_idx)

			if self.verbose == True:
				self.function.pbar.update(1)

class ApproximateLazyGreedy(BaseOptimizer):
	"""The approximate lazy/accelerated greedy algorithm for optimization.

	The approximate lazy greedy algorithm is a simple extension of the lazy 
	greedy algorithm that, rather than requiring that an element remains at 
	the top of the priority queue after being re-evaluated, only requires 
	that the gain is within a certain user-defined percentage of the best 
	gain to be selected. The key point in this approach is that finding the 
	very best element while maintaining the priority queue may be expensive, 
	but finding elements that are good enough is simple. While the best 
	percentage to use is data set specific, even values near 1 can lead to 
	large savings in computation.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='approximate-lazy')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.pq : utils.PriorityQueue
		The priority queue used to order examples for evaluation.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, beta=0.9, random_state=None, 
		n_jobs=None, verbose=False):
		self.beta = beta
		super(ApproximateLazyGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		gains = self.function._calculate_gains(X) / sample_cost[self.function.idxs]
		self.pq = PriorityQueue(self.function.idxs, -gains)

		while cost < k:
			while True:
				if len(self.pq.pq) == 0:
					return

				prev_gain, idx = self.pq.pop()
				prev_gain = -prev_gain
				
				if cost + sample_cost[idx] > k:
					continue

				idxs = numpy.array([idx])
				gain = self.function._calculate_gains(X, idxs)[0] / sample_cost[idx]

				if gain >= self.beta * prev_gain:
					best_gain = gain
					best_idx = idx
					break
				else:
					self.pq.add(idx, -gain)


			cost += sample_cost[best_idx]
			best_gain *= sample_cost[best_idx]
			self.function._select_next(X[best_idx], best_gain, best_idx)

			if self.verbose == True:
				self.function.pbar.update(1)


class TwoStageGreedy(BaseOptimizer):
	"""An approach that switches between two optimizers midway through.

	The two-stage greedy optimizer is a general purpose framework for combining 
	two optimizers by making the first :math:`n` selections out of :math:`k` 
	total selections using one optimizer, and then making the remainder using 
	the other. When the first optimizer is random selection and the second 
	approach is naive/lazy greedy, this becomes partial enumeration. By 
	default, the first algorithm is the naive greedy optimizer and the second 
	algorithm is the lazy greedy. This combination results in the same 
	selection as either optimizer individually but replaces the 
	computationally intensive first few steps for the priority queue, where 
	the algorithm may require scanning through almost the entire queue, with 
	the parallelizable naive greedy algorithm. While, in theory, the lazy 
	greedy algorithm will never perform more function calls than the naive 
	greedy algorithm, there are costs associated both with maintaining a 
	priority queue and with evaluating a single example instead of a batch 
	of examples.

	This optimizer, with the naive greedy optimizer first and the lazy greedy 
	optimizer second, is the default optimizer for apricot selectors.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='two-stage')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.n_first_selections : int
		The number of selections to perform using the naive greedy algorithm
		before populating the priority queue and using the lazy greedy
		algorithm.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.
	"""

	def __init__(self, function=None, n_first_selections=10, 
		optimizer1='naive', optimizer2='lazy', random_state=None,
		n_jobs=None, verbose=False):
		self.n_first_selections = n_first_selections
		self.optimizer1 = optimizer1
		self.optimizer2 = optimizer2
		super(TwoStageGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		if isinstance(self.optimizer1, str):
			optimizer1 = OPTIMIZERS[self.optimizer1](function=self.function,
				verbose=self.verbose)
		elif not isinstance(self.optimizer1, BaseOptimizer):
			raise ValueError("optimizer1 must be either a string or an " /
				"optimized object.")
		else:
			optimizer1 = self.optimizer1

		if isinstance(self.optimizer2, str):
			optimizer2 = OPTIMIZERS[self.optimizer2](function=self.function, 
				verbose=self.verbose)
		elif not isinstance(self.optimizer2, BaseOptimizer):
			raise ValueError("optimizer1 must be either a string or an " /
				"optimized object.")
		else:
			optimizer2 = self.optimizer2

		optimizer1.select(X, min(self.n_first_selections, k), sample_cost=sample_cost)
		if k > self.n_first_selections:
			m = k - self.n_first_selections
			optimizer2.select(X, m, sample_cost=sample_cost)


class StochasticGreedy(BaseOptimizer):
	"""The stochastic greedy algorithm for optimization.

	The stochastic greedy algorithm is a simple approach that, for each 
	iteration, randomly selects a subset of data and then finds the best next 
	example within that subset. The distinction between this approach and the 
	sample greedy algorithm is that this subset changes at each iteration, 
	meaning that the algorithm does cover the entire data set. In contrast, 
	the sample greedy algorithm is equivalent to manually subsampling the data 
	before running a selector on it. The size of this subset is proportional 
	to the number of examples that are chosen and determined in a manner that 
	results in the same amount of computation being done no matter how many 
	elements are selected. A key idea from this approach is that, while the 
	exact ranking of the elements may differ from the naive/lazy greedy 
	approaches, the set of selected elements is likely to be similar despite 
	the limited amount of computation.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='stochastic')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	epsilon : float, optional
		The inverse of the sampling probability of any particular point being 
		included in the subset, such that 1 - epsilon is the probability that
		a point is included. Default is 0.9.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, epsilon=0.9, random_state=None, 
		n_jobs=None, verbose=False):
		self.epsilon = epsilon
		super(StochasticGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		n = X.shape[0]
		subset_size = -numpy.log(self.epsilon) * n / k
		subset_size = max(int(subset_size), 1)

		while cost < k:
			idxs = self.random_state.choice(self.function.idxs, 
				replace=False, size=min(subset_size, 
					self.function.idxs.shape[0]))

			gains = self.function._calculate_gains(X, idxs) / sample_cost[idxs]
			idxs_ = numpy.lexsort((numpy.arange(gains.shape[0]), -gains))

			for idx in idxs_:
				best_idx = idxs[idx]
				if cost + sample_cost[best_idx] <= k:
					break
			else:
				return

			cost += sample_cost[best_idx]
			self.function._select_next(X[best_idx], gains[idx], best_idx)

			if self.verbose == True:
				self.function.pbar.update(1)


class SampleGreedy(BaseOptimizer):
	"""The sample greedy algorithm for optimization.

	The sample greedy algorithm is a simple approach that subsamples the full 
	data set with a user-defined sampling probability and then runs an 
	optimization on that subset. This subsampling can lead to obvious speed 
	improvements because fewer elements as selected, but will generally find 
	a lower quality subset because fewer elements are present. This approach 
	is typically used a baseline for other approaches but can save a lot of 
	time on massive data sets that are known to be highly redundant.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='sample')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	epsilon : float, optional
		The sampling probability to use when constructing the subset. A
		subset of size n * epsilon will be selected from.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, epsilon=0.9, optimizer='lazy', 
		random_state=None, n_jobs=None, verbose=False):
		self.epsilon = epsilon
		self.optimizer = optimizer
		super(SampleGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		n = X.shape[0]
		subset_size = max(int(n * self.epsilon), 1)

		idxs = self.random_state.choice(self.function.idxs,
			replace=False, size=n - subset_size)

		if isinstance(self.optimizer, str):
			optimizer = OPTIMIZERS[self.optimizer](function=self.function,
				verbose=self.verbose)
		elif not isinstance(self.optimizer, BaseOptimizer):
			raise ValueError("optimizer must be either a string or an " /
				"optimizer object.")
		else:
			optimizer = self.optimizer

		self.function.mask[idxs] = 1
		self.function.idxs = numpy.where(self.function.mask == 0)[0]
		optimizer.select(X, k, sample_cost=sample_cost)


class GreeDi(BaseOptimizer):
	"""An approach for optimizing submodular functions on massive data sets.

	GreeDi is an optimizer that was designed to work on data sets that are 
	too large to fit into memory. The approach involves first partitioning 
	the data into :math:`m` equally sized chunks without any overlap. Then, 
	:math:`l` elements are selected from each chunk using a standard optimizer 
	like naive or lazy greedy. Finally, these :math:`ml` examples are merged 
	and a standard optimizer selects :math:`k` examples from this set. In 
	this manner, the algorithm sacrifices exactness to ensure that memory 
	limitations are not an issue. 

	There are a few considerations to keep in mind when using GreeDi. 
	Naturally, :math:`ml` must both be larger than :math:`k` and also small 
	enough to fit into memory. The larger :math:`l`, the closer the solution 
	is to the exact solution but also the more compute time is required. 
	Conversely, the larger :math:`m` is, the less exact the solution is. 
	When using a graph-based function, increasing :math:`m` can dramatically 
	reduce the amount of computation that needs to be performed, as well as 
	the memory requirements, because the similarity matrix becomes smaller 
	in size. However, feature-based functions are likely to see less of a 
	speed improvement because the cost of evaluating an example is independent 
	of the size of ground set.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='greedi')
		selector.fit(X)

	Parameters
	----------
	function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	m : int
		The number of partitions to split the data into. 

	l : int
		The number of exemplars to select from each partition. l*m must
		be larger than the number of exemplars k that will be selected
		later on.

	optimizer1 : str or base.Optimizer, optional
		The optimizer to use in the first stage of the process where l
		exemplars are selected from each partition. Default is 'lazy'.

	optimizer2 : str or base.Optimizer, optional
		The optimizer to use in the second stage where k exemplars are
		selected from the l*m exemplars returned from the first stage.
		Default is 'lazy'.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process.

	verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection or None
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized. If None, will be set by the selector when passed in.

	self.m : int
		The number of partitions that the data will be split into.

	self.l : int
		The number of exemplars that will be selected from each partition.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, m=None, l=None, optimizer1='lazy', 
		optimizer2='lazy', random_state=None, n_jobs=None, verbose=False):
		self.m = m
		self.l = l
		self.optimizer1 = optimizer1
		self.optimizer2 = optimizer2
		super(GreeDi, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		if self.m is None and self.l is None:
			self.m = 8
			self.l = min(k // 4, X.shape[0] // self.m)

		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		if k > (self.m * self.l):
			raise ValueError("k must be smaller than m * l")

		partitions = numpy.arange(X.shape[0]) % self.m
		self.random_state.shuffle(partitions)

		optimizer1 = OPTIMIZERS[self.optimizer1](function=self.function, 
			verbose=self.verbose)
		optimizer2 = OPTIMIZERS[self.optimizer2](function=self.function, 
			verbose=self.verbose)

		rankings = []

		for i in range(self.m):
			idxs = numpy.where(partitions == i)[0]

			X_subset = X[idxs]
			if X.shape[0] == X.shape[1]:
				X_subset = X_subset[:, idxs]

			self.function._initialize(X_subset)
			optimizer1.select(X_subset, self.l, sample_cost=sample_cost[idxs])

			rankings.append(idxs[self.function.ranking])

		rankings = numpy.concatenate(rankings)

		if self.verbose:
			self.function.pbar.close()
			self.function.pbar = tqdm(total=k)

		X_subset = X[rankings]
		if X.shape[0] == X.shape[1]:
			X_subset = X_subset[:, rankings]

		self.function._initialize(X_subset)
		optimizer2.select(X_subset, k, sample_cost=sample_cost[rankings])
		rankings = rankings[self.function.ranking]

		self.function._initialize(X)
		for idx in rankings:
			gain = self.function._calculate_gains(X, numpy.array([idx]))[0]
			self.function._select_next(X[idx], gain, idx)


class RandomGreedy(BaseOptimizer):
	"""The naive greedy algorithm for optimization.

	This optimization approach is the naive greedy algorithm. At each iteration
	of selection it will simply calculate the gain one would get from adding
	each example, and then will select the example that has the highest gain.
	This algorithm is conceptually simple and easy to parallelize and put on a
	GPU, but can be slower than other alternatives because it involves
	repeatedly evaluating examples that are not likely to be selected next.

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.
	"""

	def __init__(self, function=None, random_state=None, n_jobs=None, 
		verbose=False):
		super(NaiveGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		i = 0
		while cost < k:
			idx = self.random_state.choice(self.random.idxs)
		
			if cost + sample_cost[idx] > k:
				continue

			cost += sample_cost[idx]
			gain = self.function._calculate_gains(X, idxs=numpy.array([idx]))[0]

			self.function._select_next(X[idx], gain, idx)

			if self.verbose == True:
				self.function.pbar.update(round(sample_cost[idx], 2))


class ModularGreedy(BaseOptimizer):
	"""An approach that approximates a submodular function as modular.

	This approach approximates the submodular function by using its modular 
	upper-bounds to do the selection. Essentially, a defining characteristic 
	of submodular functions is the *diminishing returns* property where the 
	gain of an example decreases with the number of selected examples. In 
	contrast, modular functions have constant gains for examples regardless 
	of the number of selected examples. Thus, approximating the submodular 
	function as a modular function can serve as an upper-bound to the gain 
	for each example during the selection process. This approximation makes 
	the function simple to optimize because one would simply calculate the 
	gain that each example yields before any examples are selected, sort the 
	examples by this gain, and select the top :math:`k` examples. While this 
	approach is fast, this approach is likely best paired with a traditional 
	optimization algorithm after the first few examples are selected.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='modular')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	epsilon : float, optional
		The inverse of the sampling probability of any particular point being 
		included in the subset, such that 1 - epsilon is the probability that
		a point is included. Default is 0.9.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, random_state=None, n_jobs=None, 
		verbose=False):
		super(ModularGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		"""Select elements in a naive greedy manner."""

		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		gains = self.function._calculate_gains(X) / sample_cost[self.function.idxs]
		idxs = gains.argsort()[::-1]

		for idx in idxs:
			if cost + sample_cost[idx] > k:
				continue

			cost += sample_cost[idx]
			gain = self.function._calculate_gains(X, idxs=numpy.array([idx]))[0]

			self.function._select_next(X[idx], gain, idx)

			if self.verbose == True:
				self.function.pbar.update(1)


class BidirectionalGreedy(BaseOptimizer):
	"""The bidirectional greedy algorithm.

	Most submodular optimizers assume that the function is *monotone*, i.e., 
	that the gain from each successive example is positive. However, there 
	are some cases where the key diminishing returns property holds, but 
	the gains are not necessarily positive. The most obvious of these is a
	difference in submodular functions. In these cases, the naive greedy 
	algorithm is not guaranteed to return a good result. 

	The bidirectional greedy algorithm was developed to optimize non-monotone 
	submodular functions. While it has a guarantee that is lower than the 
	naive greedy algorithm has for monotone functions, it generally returns 
	better sets than the greedy algorithm.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='bidirectional')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, random_state=None, n_jobs=None, 
		verbose=False):
		super(BidirectionalGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		"""Select elements in a naive greedy manner."""

		cost = 0.0
		if sample_cost is None:
			sample_cost = numpy.ones(X.shape[0], dtype='float64')

		A = numpy.zeros(X.shape[0], dtype=bool)
		B = numpy.ones(X.shape[0], dtype=bool)

		idxs = numpy.arange(X.shape[0])
		self.random_state.shuffle(idxs)
		gains = numpy.zeros(X.shape[0])

		while True: 
			for i in idxs:
				if A[i] == True:
					continue

				self.function.initial_subset = A
				self.function._initialize(X)
				gain_a = self.function._calculate_gains(X[i])


				B[i] = False
				self.function.initial_subset = B
				self.function._initialize(X)
				gain_b = -self.function._calculate_gains(X[i])
				B[i] = True

				if gain_a == gain_b == 0.0:
					p = 0.5
				else:
					p = gain_a / (gain_a + gain_b)

				self.gains_[i] = p

				if self.random_state.uniform(0, 1) <= p:
					A[i] = True
					self.function._select_next(X[i], gain_a, i)
					if A.sum() == n:
						return

				else:
					B[i] = False


class SieveGreedy(BaseOptimizer):
	"""The sieve stream greedy algorithm.

	Most submodular optimizers assume that the function is *monotone*, i.e., 
	that the gain from each successive example is positive. However, there 
	are some cases where the key diminishing returns property holds, but 
	the gains are not necessarily positive. The most obvious of these is a
	difference in submodular functions. In these cases, the naive greedy 
	algorithm is not guaranteed to return a good result. 

	The bidirectional greedy algorithm was developed to optimize non-monotone 
	submodular functions. While it has a guarantee that is lower than the 
	naive greedy algorithm has for monotone functions, it generally returns 
	better sets than the greedy algorithm.

	.. code::python

		from apricot import FeatureBasedSelection

		X = numpy.random.randint(10, size=(10000, 100))

		selector = FeatureBasedSelection(100, 'sqrt', optimizer='bidirectional')
		selector.fit(X)

	Parameters
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.


	Attributes
	----------
	self.function : base.BaseSelection
		A submodular function that implements the `_calculate_gains` and
		`_select_next` methods. This is the function that will be
		optimized.

	self.verbose : bool
		Whether to display a progress bar during the optimization process.

	self.gains_ : numpy.ndarray or None
		The gain that each example would give the last time that it was
		evaluated.
	"""

	def __init__(self, function=None, epsilon=0.01, random_state=None, 
		n_jobs=None, verbose=False):
		self.epsilon = epsilon
		self.n_seen_ = 0
		self.thresholds = [1]
		self.max_gain = -1

		super(SieveGreedy, self).__init__(function=function, 
			random_state=random_state, n_jobs=n_jobs, verbose=verbose)

	def select(self, X, k, sample_cost=None):
		"""Select elements in a naive greedy manner."""

		# This is not a great use of the apricot API. An optimizer shouldn't
		# be using its own special gain calculating function. However, it is
		# much faster to do things this way so I'll keep it in for now.

		n, d = X.shape
		if sample_cost is None:
			sample_cost = numpy.ones(n, dtype='float64')

		r = X.shape[1] if self.function.reservoir is not None else 1
		marginal_gains = self.function._calculate_gains(X) / sample_cost / r
		max_marginal_gain = marginal_gains.max()
		if max_marginal_gain > self.max_gain:
			self.max_gain = max_marginal_gain
			j = len(self.thresholds)
			threshold = (1 + self.epsilon) ** j - 1
			while threshold <= self.max_gain * k:
				self.thresholds.append(threshold)

				j += 1
				threshold = (1 + self.epsilon) ** j - 1

		thresholds = numpy.array(self.thresholds, dtype='float64')
		idxs = numpy.arange(n, dtype='int64') + self.n_seen_

		self.function._calculate_sieve_gains(X, thresholds, idxs)

		best_idx = numpy.argmax(self.function.sieve_total_gains_)
		ranking = self.function.sieve_selections_[best_idx]
		ranking = ranking[:self.function.sieve_n_selected_[best_idx]]
		gain = self.function.sieve_gains_[best_idx]
		gain = gain[:self.function.sieve_n_selected_[best_idx]]

		for i in range(len(self.thresholds)):
			for j in range(k):
				m = self.function.sieve_selections_[i, j]
				if m >= self.n_seen_:
					if isinstance(self.function._X, scipy.sparse.csr_matrix):
						self.function.sieve_subsets_[i, j] = self.function._X[m - self.n_seen_].toarray()[0]
					else:
						self.function.sieve_subsets_[i, j] = self.function._X[m - self.n_seen_]

		self.function.ranking = ranking
		self.function.gains = gain
		self.function.subset = self.function.sieve_subsets_[best_idx]
		self.n_seen_ += X.shape[0]


OPTIMIZERS = {
	'random' : RandomGreedy,
	'modular' : ModularGreedy,
	'naive' : NaiveGreedy,
	'lazy' : LazyGreedy,
	'approximate-lazy' : ApproximateLazyGreedy,
	'two-stage' : TwoStageGreedy,
	'stochastic' : StochasticGreedy,
	'sample' : SampleGreedy,
	'greedi' : GreeDi,
	'bidirectional' : BidirectionalGreedy,
	'sieve' : SieveGreedy,
}



# utils.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

"""
This code contains utility functions to support the main functionality of
the code.
"""

import numbers
import numpy
import itertools

from heapq import heappush
from heapq import heappop
from heapq import heapify

from scipy.sparse import csr_matrix

from sklearn.metrics import pairwise_distances
from sklearn.neighbors import KNeighborsTransformer

class PriorityQueue(object):
    """A priority queue implementation.

    This is an implemented of a priority queue using a heap implemtnation. It
    has O(log n) time for adding and popping elements but constant time
    deletion. It is based off of the implementation that is provided at

    https://docs.python.org/2/library/heapq.html

    This implementation assumes that the items with the highest priority
    are the ones with the lowest "weight" that gets passed in. If higher
    weights are supposed to correspond to higher priority, consider reversing
    the sign of the weight.

    Parameters
    ----------
    None

    Attributes
    ----------
    pq : list
        A list containing the heapified priority queue.

    lookup : dict
        A dictionary containing links to each of the elements, allowing
        constant time lookups and deletions.
    """

    def __init__(self, items=None, weights=None):
        self.counter = itertools.count()
        self.lookup = {}
        self.pq = []

        if items is not None and weights is not None:
            for item, weight in zip(items, weights):
                entry = [weight, next(self.counter), item]
                self.pq.append(entry)
                self.lookup[item] = entry

            heapify(self.pq)
    
    def add(self, item, weight):
        """Add an element to the priority queue. Runtime is O(log n).

        This adds in an element to the priority queue. If that element is
        already there it will automatically delete the previous instance
        and replace it with the new instance.

        Parameters
        ----------
        item : object
            The object to be encoded.

        weight : double
            The priority of the item. The lower the weight the higher the
            priority when items get dequeued. If a higher weight is supposed
            to correspond to a higher priority, consider reversing the sign of
            the weight.

        Returns
        -------
        None
        """

        if item in self.lookup:
            self.remove(item)
        
        entry = [weight, next(self.counter), item]
        self.lookup[item] = entry
        heappush(self.pq, entry)

    def remove(self, item):
        """Remove an element from the queue.

        This is not popping the highest priority item, rather it will remove
        an element that is present in the queue. If one attempts to remove an
        item that is not present in the queue the function will error.
        
        Parameters
        ----------
        item : object
            The object to be removed from the queue.

        Returns
        -------
        None
        """

        entry = self.lookup.pop(item)
        entry[-1] = "DELETED"
    
    def pop(self):
        """Pop the highest priority element from the queue. Runtime is O(log n).

        This will remove the highest priority element from the queue. If there
        are no elements left in the queue it will raise an error.

        Parameters
        ----------
        None

        Returns
        -------
        weight : double
            The weight of the element as passed in in the `add` method

        item : object
            The item that was passed in during the `add` method
        """

        while self.pq:
            weight, _, item = heappop(self.pq)
            if item != "DELETED":
                del self.lookup[item]
                return weight, item
        
        raise KeyError("No elements left in the priority queue.")

def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance.

    This function will check to see whether the input seed is a valid seed
    for generating random numbers. This is a slightly modified version of
    the code from sklearn.utils.validation.

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    """

    if seed is None or seed is numpy.random:
        return numpy.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, numpy.integer)):
        return numpy.random.RandomState(seed)
    if isinstance(seed, numpy.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)

def _calculate_pairwise_distances(X, Y=None, metric='precomputed', n_neighbors=None):
    if metric in ('precomputed', 'ignore'):
        return X

    if n_neighbors is None:
        if metric == 'euclidean':
            X_pairwise = pairwise_distances(X, Y=Y, metric=metric, squared=True)
        elif metric == 'correlation' or metric == 'cosine':
            # An in-place version of:
            # X_pairwise = 1 - (1 - pairwise_distances(X, metric=metric)) ** 2
            
            X_pairwise = pairwise_distances(X, Y=Y, metric=metric)
            X_pairwise = numpy.subtract(1, X_pairwise, out=X_pairwise)
            X_pairwise = numpy.square(X_pairwise, out=X_pairwise)
            X_pairwise = numpy.subtract(1, X_pairwise, out=X_pairwise)
        else:
            X_pairwise = pairwise_distances(X, Y=Y, metric=metric)
    else:
        if metric == 'correlation' or metric == 'cosine':
            # An in-place version of:
            # X = 1 - (1 - pairwise_distances(X, metric=metric)) ** 2

            X = pairwise_distances(X, Y=Y, metric=metric)
            X = numpy.subtract(1, X, out=X)
            X = numpy.square(X, out=X)
            X = numpy.subtract(1, X, out=X)
            metric = 'precomputed'

        if isinstance(n_neighbors, int):
            X_pairwise = KNeighborsTransformer(
                n_neighbors=n_neighbors, metric=metric
                ).fit_transform(X)

        elif isinstance(n_neighbors, KNeighborsTransformer):
            X_pairwise = n_neighbors.fit_transform(X)

    if metric == 'correlation' or metric == 'cosine':
        if isinstance(X_pairwise, csr_matrix):
            X_pairwise.data = numpy.subtract(1, X_pairwise.data, 
                out=X_pairwise.data)
        else:
            X_pairwise = numpy.subtract(1, X_pairwise,
                out=X_pairwise)
    else:
        if isinstance(X_pairwise, csr_matrix):
            X_pairwise.data = numpy.subtract(X_pairwise.max(),
                X_pairwise.data, out=X_pairwise.data)
        else:
            X_pairwise = numpy.subtract(X_pairwise.max(), X_pairwise,
                out=X_pairwise)

    return X_pairwise


# maxCoverage.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com> 
	
import numpy

from .base import BaseSelection
from ..optimizers import LazyGreedy
from ..optimizers import ApproximateLazyGreedy
from ..optimizers import SieveGreedy

from numba import njit
from numba import prange

dtypes = 'void(float64[:,:], float64[:], float64[:], float64, int64[:])'
sdtypes = 'void(float64[:], int32[:], int32[:], float64[:], float64[:], float64, int64[:])'
sieve_dtypes = 'void(float64[:,:], int64, float64[:,:], int64[:,:],' \
	'float64[:,:], float64[:], float64[:], int64[:], float64, int64[:])'
sieve_sparse_dtypes = 'void(float64[:], int32[:], int32[:], int64,' \
	'float64[:,:], int64[:,:], float64[:,:], float64[:], float64[:],' \
	'int64[:], float64, int64[:])'

def calculate_gains(dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_(X, gains, current_values, threshold, idxs):
		for i in prange(idxs.shape[0]):
			idx = idxs[i] 
			gains[i] = numpy.fmin(current_values + X[idx], threshold).sum()
	return calculate_gains_


def calculate_gains_sparse(dtypes, parallel, fastmath, cache):
	@njit(sdtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_sparse_(X_data, X_indices, X_indptr, gains, 
		current_values, threshold, idxs):
		for i in prange(idxs.shape[0]):
			idx = idxs[i]
			start = X_indptr[idx]
			end = X_indptr[idx+1]

			for j in range(start, end):
				k = X_indices[j]
				gains[i] += min(X_data[j] + current_values[k], threshold) - current_values[k]
	return calculate_gains_sparse_


def calculate_gains_sieve(dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_sieve_(X, k, current_values, selections, gains, 
		total_gains, max_values, n_selected, thresh, idxs):
		n = X.shape[0]
		d = max_values.shape[0]

		for j in prange(d):
			for i in range(n):
				if n_selected[j] == k:
					break

				idx = idxs[i]
				threshold = (max_values[j] / 2. - total_gains[j]) / (k - n_selected[j])
				gain = (numpy.fmin(current_values[j] + X[i], thresh).sum() - 
					total_gains[j])

				if gain > threshold:
					current_values[j] += X[i]
					total_gains[j] += gain

					selections[j, n_selected[j]] = idx
					gains[j, n_selected[j]] = gain
					n_selected[j] += 1

	return calculate_gains_sieve_


def calculate_gains_sieve_sparse(dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_sieve_sparse_(X_data, X_indices, X_indptr, k, 
		current_values, selections, gains, total_gains, max_values, 
		n_selected, thresh, idxs):
		d = max_values.shape[0]

		for j in prange(d):
			for i in range(idxs.shape[0]):
				if n_selected[j] == k:
					break
				
				idx = idxs[i]
				start = X_indptr[i]
				end = X_indptr[i+1]
				threshold = (max_values[j] / 2. - total_gains[j]) / (k - n_selected[j])

				gain = 0.0
				for l in range(start, end):
					m = X_indices[l]
					gain += (min(current_values[j, m] + X_data[l], thresh) - 
						current_values[j, m])

				if gain > threshold:
					for l in range(start, end):
						m = X_indices[l]
						current_values[j, m] = min(current_values[j, m] + X_data[l], 
							thresh)

					total_gains[j] += gain
					selections[j, n_selected[j]] = idx
					gains[j, n_selected[j]] = gain
					n_selected[j] += 1

	return calculate_gains_sieve_sparse_

class MaxCoverageSelection(BaseSelection):
	"""A selector based off a coverage function.

	Maximum coverage functions aim to maximize the number of features that 
	have a non-zero element in at least one selected example---there is no 
	marginal benefit to observing a variable in two examples. If each variable 
	is thought to be an item in a set, and the data is a binary matrix where a 
	1 indicates the item is present in the example and 0 indicates it is not, 
	optimizing a maximum coverage function is a solution to the set coverage 
	problem. These functions are useful when the space of variables is massive 
	and each example only sees a small subset of them, which is a common 
	situation when analyzing text data when the variables are words. The 
	maximum coverage function is an instance of a feature-based function 
	when the concave function is minimum.

	.. note::
		All values in your data must be binary for this selection to work.

	The general form of a coverage function is:

	.. math::
		f(X) = \\sum\\limits_{d=1}^{D} \\min \\left( \\sum\\limits_{n=1}^{N} X_{i, d}, 1 \\right) 

	where :math:`f` indicates the function that operates on a subset :math:`X` 
	that has :math:`N` examples and :math:`D` dimensions. Importantly, 
	:math:`X` is the subset and not the ground set, meaning that the time it 
	takes to evaluate this function is proportional only to the size of the 
	selected subset and not the size of the full data set, like it is for
	 graph-based functions.  

	See https://www2.cs.duke.edu/courses/fall17/compsci632/scribing/scribe2.pdf
	where the problem is described as maximum coverage.

	Parameters
	----------
	n_samples : int
		The number of examples to return.

	initial_subset : list, numpy.ndarray or None
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'random' : randomly select elements (dummy optimizer)
			'modular' : approximate the function using its modular upper bound
			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'sample' : randomly take a subset and perform selection on that
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'two-stage'.

	optimizer_kwds : dict, optional
		Arguments to pass into the optimizer object upon initialization.
		Default is {}.

	n_jobs : int, optional
		The number of cores to use for processing. This value is multiplied
		by 2 when used to set the number of threads. If set to -1, use all
		cores and threads. Default is -1.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, threshold=1.0, initial_subset=None, 
		optimizer='two-stage', optimizer_kwds={}, n_jobs=1, random_state=None, 
		verbose=False):
		self.threshold = threshold

		super(MaxCoverageSelection, self).__init__(n_samples=n_samples, 
			initial_subset=initial_subset, optimizer=optimizer, 
			optimizer_kwds=optimizer_kwds, n_jobs=n_jobs, 
			random_state=random_state, verbose=verbose) 

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select the examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : FeatureBasedSelection
			The fit step returns this selector object.
		"""

		return super(MaxCoverageSelection, self).fit(X, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

	def _initialize(self, X):
		super(MaxCoverageSelection, self)._initialize(X)

		if self.initial_subset is None:
			self.current_values = numpy.zeros(X.shape[1], dtype='float64')
		elif self.initial_subset.ndim == 2:
			self.current_values = numpy.fmin(self.threshold,
				self.initial_subset.sum(axis=0).astype('float64'))
		elif self.initial_subset.ndim == 1:
			self.current_values = numpy.fmin(self.threshold,
				X[self.initial_subset].sum(axis=0).astype('float64'))
		else:
			raise ValueError("The initial subset must be either a two dimensional" \
				" matrix of examples or a one dimensional mask.")

		self.current_values_sum = self.current_values.sum() 

		calculate_gains_ = calculate_gains_sparse if self.sparse else calculate_gains
		dtypes_ = sdtypes if self.sparse else dtypes

		if self.optimizer in (LazyGreedy, ApproximateLazyGreedy):
			self.calculate_gains_ = calculate_gains_(dtypes_, False, True, False)
		elif self.optimizer in ('lazy', 'approimate-lazy'):
			self.calculate_gains_ = calculate_gains_(dtypes_, False, True, False)
		else: 
			self.calculate_gains_ = calculate_gains_(dtypes_, True, True, False)

		calculate_sieve_gains_ = calculate_gains_sieve_sparse if self.sparse else calculate_gains_sieve
		dtypes_ = sieve_sparse_dtypes if self.sparse else sieve_dtypes 
		self.calculate_sieve_gains_ = calculate_sieve_gains_(dtypes_, True, 
			True, False)

	def _calculate_gains(self, X, idxs=None):
		"""This function will return the gain that each example would give.

		This function will return the gains that each example would give if
		added to the selected set. When a matrix of examples is given, a
		vector will be returned showing the gain for each example. When
		a single element is passed in, it will return a singe value."""

		idxs = idxs if idxs is not None else self.idxs
		gains = numpy.zeros(idxs.shape[0], dtype='float64')

		if self.sparse:
			self.calculate_gains_(X.data, X.indices, X.indptr, gains, 
				self.current_values, self.threshold, idxs)
		else:
			self.calculate_gains_(X, gains, self.current_values, 
				self.threshold, idxs)
			gains -= self.current_values_sum

		return gains

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		"""This function will update the internal statistics from a stream.

		This function will update the various internal statistics that are a
		part of the sieve algorithm for streaming submodular optimization. This
		function does not directly return gains but it updates the values
		used by a streaming optimizer.
		"""

		super(MaxCoverageSelection, self)._calculate_sieve_gains(X,
			thresholds, idxs)

		if self.sparse:
			self.calculate_sieve_gains_(X.data, X.indices, X.indptr, 
				self.n_samples, self.sieve_current_values_, 
				self.sieve_selections_, self.sieve_gains_, 
				self.sieve_total_gains_, thresholds, 
				self.sieve_n_selected_, self.threshold, idxs)
		else:
			self.calculate_sieve_gains_(X, self.n_samples, 
				self.sieve_current_values_, self.sieve_selections_, 
				self.sieve_gains_, self.sieve_total_gains_, thresholds, 
				self.sieve_n_selected_, self.threshold, idxs)

	def _select_next(self, X, gain, idx):
		"""This function will add the given item to the selected set."""

		if self.sparse:
			self.current_values = numpy.fmin(self.threshold, 
				self.current_values + X.toarray()[0])
		else:
			self.current_values = numpy.fmin(self.threshold, 
				self.current_values + X)

		self.current_values_sum = self.current_values.sum()

		super(MaxCoverageSelection, self)._select_next(
			X, gain, idx)


# sumRedundancy.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy

from .base import BaseGraphSelection

from tqdm import tqdm

class SumRedundancySelection(BaseGraphSelection):
	"""A selector based off a sum redundancy submodular function.
	
	The sum redundancy function is a graph-based function that penalizes 
	redundancy among the selected set. This approach is straightforward, 
	in that it simply involved a sum. It is also fast in comparison to a 
	facility location function because it involves only performing calculation 
	over the selected set as opposed to the entire ground set. Because the sum 
	of the similarities is not submodular, it is subtracted from the sum of 
	the entire similarity matrix, such that examples that are highly similar 
	to each other result in a lower value than examples that are not very 
	similar.

	.. note:: 
		All ~pairwise~ values in your data must be positive for this 
		selection to work.

	The general form of a sum redundancy function is 

	.. math::
		f(X, V) = \sum_{x, y \in V} \phi(x, y) - \sum_{x, y\in X} \phi(x,y)

	where :math:`f` indicates the function, :math:`X` is the selected subset, 
	:math:`V` is the ground set, and :math:`\phi` is the similarity measure 
	between two examples. While sum redundancy functions involves calculating 
	the sum of the entire similarity matrix in principle, in practice if one 
	is only calculating the gains this step can be ignored.
	
	This implementation allows users to pass in either their own symmetric
	square matrix of similarity values, or a data matrix as normal and a function
	that calculates these pairwise values.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	metric : str, optional
		The method for converting a data matrix into a square symmetric matrix
		of pairwise similarities. If a string, can be any of the metrics
		implemented in sklearn (see https://scikit-learn.org/stable/modules/
		generated/sklearn.metrics.pairwise_distances.html), including
		"precomputed" if one has already generated a similarity matrix. Note
		that sklearn calculates distance matrices whereas apricot operates on
		similarity matrices, and so a distances.max() - distances transformation
		is performed on the resulting distances. For backcompatibility,
		'corr' will be read as 'correlation'. Default is 'euclidean'.


	initial_subset : list, numpy.ndarray or None, optional
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional. Default is None.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'random' : randomly select elements (dummy optimizer)
			'modular' : approximate the function using its modular upper bound
			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'sample' : randomly take a subset and perform selection on that
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'two-stage'.

	optimizer_kwds : dict or None
		A dictionary of arguments to pass into the optimizer object. The keys
		of this dictionary should be the names of the parameters in the optimizer
		and the values in the dictionary should be the values that these
		parameters take. Default is None.

	n_neighbors : int or None
		When constructing a similarity matrix, the number of nearest neighbors
		whose similarity values will be kept. The result is a sparse similarity
		matrix which can significantly speed up computation at the cost of
		accuracy. Default is None.

	reservoir : numpy.ndarray or None
		The reservoir to use when calculating gains in the sieve greedy
		streaming optimization algorithm in the `partial_fit` method.
		Currently only used for graph-based functions. If a numpy array
		is passed in, it will be used as the reservoir. If None is passed in,
		will use reservoir sampling to collect a reservoir. Default is None.

	max_reservoir_size : int 
		The maximum size that the reservoir can take. If a reservoir is passed
		in, this value is set to the size of that array. Default is 1000.

	n_jobs : int
		The number of threads to use when performing computation in parallel.
		Currently, this parameter is exposed but does not actually do anything.
		This will be fixed soon.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	pairwise_func : callable
		A function that takes in a data matrix and converts it to a square
		symmetric matrix.

	ranking : numpy.array int
		The selected samples in the order of their gain.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples=10, metric='euclidean', 
		initial_subset=None, optimizer='two-stage', n_neighbors=None, 
		reservoir=None, max_reservoir_size=1000, n_jobs=1, 
		random_state=None, optimizer_kwds={}, verbose=False):

		super(SumRedundancySelection, self).__init__(n_samples=n_samples, 
			metric=metric, initial_subset=initial_subset, optimizer=optimizer, 
			n_neighbors=n_neighbors, reservoir=reservoir, 
			max_reservoir_size=max_reservoir_size, n_jobs=n_jobs, 
			random_state=random_state, optimizer_kwds=optimizer_kwds, 
			verbose=verbose)

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select the examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : SumRedundancySelection
			The fit step returns this selector object.
		"""

		return super(SumRedundancySelection, self).fit(X, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

	def _initialize(self, X_pairwise, idxs=None):
		super(SumRedundancySelection, self)._initialize(X_pairwise, idxs=idxs)
		idxs = idxs if idxs is not None else numpy.arange(X_pairwise.shape[0])

		for i, idx in enumerate(idxs):
			self.current_values[i] = X_pairwise[idx, idx]

		if self.initial_subset is None:
			return
		elif self.initial_subset.ndim == 2:
			raise ValueError("When using saturated coverage, the initial subset"\
				" must be a one dimensional array of indices.")
		elif self.initial_subset.ndim == 1:
			if not self.sparse:
				for i in self.initial_subset:
					self.current_values += X_pairwise[i] * 2 
			else:
				for i in self.initial_subset:
					self.current_values += X_pairwise[i].toarray()[0] * 2
		else:
			raise ValueError("The initial subset must be either a two dimensional" \
				" matrix of examples or a one dimensional mask.")

	def _calculate_gains(self, X_pairwise, idxs=None):
		idxs = idxs if idxs is not None else self.idxs
		return -self.current_values[idxs]

	def _select_next(self, X_pairwise, gain, idx):
		"""This function will add the given item to the selected set."""

		if self.sparse:
			self.current_values += X_pairwise.toarray()[0] * 2
		else:
			self.current_values += X_pairwise * 2

		super(SumRedundancySelection, self)._select_next(
			X_pairwise, gain, idx)



from setuptools import setup

setup(
    name='apricot-select',
    version='0.6.0',
    author='Jacob Schreiber',
    author_email='jmschreiber91@gmail.com',
    packages=['apricot', 'apricot/functions'],
    url='http://pypi.python.org/pypi/apricot-select/',
    license='LICENSE.txt',
    description='apricot is a package for submodular selection of representative sets for machine learning models.',
    install_requires=[
        "numpy >= 1.14.2",
        "scipy >= 1.0.0",
        "numba >= 0.43.0",
        "tqdm >= 4.24.0",
        "nose"
    ],
)



# featureBased.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com> 
	
import math
import numpy

from .base import BaseSelection
from ..optimizers import LazyGreedy
from ..optimizers import ApproximateLazyGreedy

from tqdm import tqdm

from numba import njit
from numba import prange

@njit('float64[:](float64[:])', fastmath=True)
def sigmoid(X):
	return X / (1. + X)

dtypes = 'void(float64[:,:], float64[:], float64[:], int64[:])'
sparse_dtypes = 'void(float64[:], int32[:], int32[:], float64[:],' \
	'float64[:], float64[:], int64[:])'
sieve_dtypes = 'void(float64[:,:], int64, float64[:,:], int64[:,:],' \
	'float64[:,:], float64[:], float64[:], int64[:], int64[:])' 
sieve_sparse_dtypes = 'void(float64[:], int32[:], int32[:], int64,' \
	'float64[:,:], int64[:,:], float64[:,:], float64[:], float64[:],' \
	'int64[:], int64[:])'

def calculate_gains(func, dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_(X, gains, current_values, idxs):
		for i in prange(idxs.shape[0]):
			idx = idxs[i] 
			gains[i] = func(current_values + X[idx]).sum()

	return calculate_gains_


def calculate_gains_sparse(func, dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_sparse_(X_data, X_indices, X_indptr, gains, 
		current_values, current_concave_values, idxs):
		for i in prange(idxs.shape[0]):
			idx = idxs[i]
			start = X_indptr[idx]
			end = X_indptr[idx+1]

			for j in range(start, end):
				k = X_indices[j]
				gains[i] += (func(X_data[j] + current_values[k]) 
					- current_concave_values[k])

	return calculate_gains_sparse_


def calculate_gains_sieve(func, dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_sieve_(X, k, current_values, selections, gains, 
		total_gains, max_values, n_selected, idxs):
		n = X.shape[0]
		d = max_values.shape[0]

		for j in prange(d):
			for i in range(n):
				if n_selected[j] == k:
					break

				idx = idxs[i]
				threshold = (max_values[j] / 2. - total_gains[j]) / (k - n_selected[j])
				gain = func(current_values[j] + X[i]).sum() - total_gains[j]

				if gain > threshold:
					current_values[j] += X[i]
					total_gains[j] += gain

					selections[j, n_selected[j]] = idx
					gains[j, n_selected[j]] = gain
					n_selected[j] += 1

	return calculate_gains_sieve_


def calculate_gains_sieve_sparse(func, dtypes, parallel, fastmath, cache):
	@njit(dtypes, parallel=parallel, fastmath=fastmath, cache=cache)
	def calculate_gains_sieve_sparse_(X_data, X_indices, X_indptr, k, 
		current_values, selections, gains, total_gains, max_values, 
		n_selected, idxs):
		d = max_values.shape[0]

		for j in prange(d):
			for i in range(idxs.shape[0]):
				if n_selected[j] == k:
					break
				
				idx = idxs[i]
				start = X_indptr[i]
				end = X_indptr[i+1]
				threshold = (max_values[j] / 2. - total_gains[j]) / (k - n_selected[j])

				gain = 0.0
				for l in range(start, end):
					m = X_indices[l]
					gain += (func(current_values[j, m] + X_data[l]) - 
						func(current_values[j, m]))

				if gain > threshold:
					for l in range(start, end):
						m = X_indices[l]
						current_values[j, m] += X_data[l]

					total_gains[j] += gain
					selections[j, n_selected[j]] = idx
					gains[j, n_selected[j]] = gain
					n_selected[j] += 1

	return calculate_gains_sieve_sparse_


class FeatureBasedSelection(BaseSelection):
	"""A selector based off a feature based submodular function.

	Feature-based functions are those that operate on the feature values of 
	examples directly, rather than on a similarity matrix (or graph) derived 
	from those features, as graph-based functions do. Because these functions 
	do not require calculating and storing a :math:`\\mathcal{O}(n^{2})` sized 
	matrix of similarities, they can easily scale to data sets with millions of 
	examples. 
	
	.. note:: 
		All values in your data must be positive for this selection to work.

	The general form of a feature-based function is:

	.. math::
		f(X) = \\sum\\limits_{d=1}^{D}\\phi\\left(\\sum\\limits_{i=1}^{N} X_{i, d}\\right)

	where :math:`f` indicates the function that uses the concave function 
	:math:`\\phi` and is operating on a subset :math:`X` that has :math:`N` 
	examples and :math:`D` dimensions. Importantly, :math:`X` is the subset 
	and not the ground set, meaning that the time it takes to evaluate this 
	function is proportional only to the size of the selected subset and not 
	the size of the full data set, like it is for graph-based functions.  

	See https://ieeexplore.ieee.org/document/6854213 for more details on
	feature based functions.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	concave_func : str or callable
		The type of concave function to apply to the feature values. You can
		pass in your own function to apply. Otherwise must be one of the
		following:

			'log' : log(1 + X)
			'sqrt' : sqrt(X)
			'sigmoid' : X / (1 + X)

	initial_subset : list, numpy.ndarray or None
		If provided, this should be a list of indices into the data matrix
		to use as the initial subset, or a group of examples that may not be
		in the provided data should beused as the initial subset. If indices, 
		the provided array should be one-dimensional. If a group of examples,
		the data should be 2 dimensional.

	optimizer : string or optimizers.BaseOptimizer, optional
		The optimization approach to use for the selection. Default is
		'two-stage', which makes selections using the naive greedy algorithm
		initially and then switches to the lazy greedy algorithm. Must be
		one of

			'random' : randomly select elements (dummy optimizer)
			'modular' : approximate the function using its modular upper bound
			'naive' : the naive greedy algorithm
			'lazy' : the lazy (or accelerated) greedy algorithm
			'approximate-lazy' : the approximate lazy greedy algorithm
			'two-stage' : starts with naive and switches to lazy
			'stochastic' : the stochastic greedy algorithm
			'sample' : randomly take a subset and perform selection on that
			'greedi' : the GreeDi distributed algorithm
			'bidirectional' : the bidirectional greedy algorithm

		Default is 'two-stage'.

	optimizer_kwds : dict or None
		A dictionary of arguments to pass into the optimizer object. The keys
		of this dictionary should be the names of the parameters in the optimizer
		and the values in the dictionary should be the values that these
		parameters take. Default is None.

	reservoir : numpy.ndarray or None
		The reservoir to use when calculating gains in the sieve greedy
		streaming optimization algorithm in the `partial_fit` method.
		Currently only used for graph-based functions. If a numpy array
		is passed in, it will be used as the reservoir. If None is passed in,
		will use reservoir sampling to collect a reservoir. Default is None.

	max_reservoir_size : int 
		The maximum size that the reservoir can take. If a reservoir is passed
		in, this value is set to the size of that array. Default is 1000.

	n_jobs : int
		The number of threads to use when performing computation in parallel.
		Currently, this parameter is exposed but does not actually do anything.
		This will be fixed soon.

	random_state : int or RandomState or None, optional
		The random seed to use for the random selection process. Only used
		for stochastic greedy.

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.

	gains : numpy.array float
		The gain of each sample in the returned set when it was added to the
		growing subset. The first number corresponds to the gain of the first
		added sample, the second corresponds to the gain of the second added
		sample, and so forth.
	"""

	def __init__(self, n_samples, concave_func='sqrt', initial_subset=None, 
		optimizer='two-stage', optimizer_kwds={}, reservoir=None,
		max_reservoir_size=1000, n_jobs=1, random_state=None, 
		verbose=False):
		self.concave_func_name = concave_func

		if concave_func == 'log':
			self.concave_func = numpy.log1p
		elif concave_func == 'sqrt':
			self.concave_func = numpy.sqrt
		elif concave_func == 'sigmoid':
			self.concave_func = sigmoid
		elif callable(concave_func):
			self.concave_func = concave_func
		else:
			raise KeyError("Must be one of 'log', 'sqrt', 'sigmoid', or a any other numpy, scipy, or numba-compiled function.")

		super(FeatureBasedSelection, self).__init__(n_samples=n_samples, 
			initial_subset=initial_subset, optimizer=optimizer, 
			optimizer_kwds=optimizer_kwds, reservoir=reservoir,
			max_reservoir_size=max_reservoir_size, n_jobs=n_jobs, 
			random_state=random_state, verbose=verbose) 

	def fit(self, X, y=None, sample_weight=None, sample_cost=None):
		"""Run submodular optimization to select the examples.

		This method is a wrapper for the full submodular optimization process.
		It takes in some data set (and optionally labels that are ignored
		during this process) and selects `n_samples` from it in the greedy
		manner specified by the optimizer.

		This method will return the selector object itself, not the transformed
		data set. The `transform` method will then transform a data set to the
		selected points, or alternatively one can use the ranking stored in
		the `self.ranking` attribute. The `fit_transform` method will perform
		both optimization and selection and return the selected items.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray or None, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		sample_weight : list or numpy.ndarray or None, shape=(n,), optional
			The weight of each example. Currently ignored in apricot but
			included to maintain compatibility with sklearn pipelines. 

		sample_cost : list or numpy.ndarray or None, shape=(n,), optional
			The cost of each item. If set, indicates that optimization should
			be performed with respect to a knapsack constraint.

		Returns
		-------
		self : FeatureBasedSelection
			The fit step returns this selector object.
		"""

		return super(FeatureBasedSelection, self).fit(X, y=y, 
			sample_weight=sample_weight, sample_cost=sample_cost)

	def _initialize(self, X):
		super(FeatureBasedSelection, self)._initialize(X)

		if self.initial_subset is None:
			pass
		elif self.initial_subset.ndim == 2:
			self.current_values = self.initial_subset.sum(axis=0).astype('float64')
		elif self.initial_subset.ndim == 1:
			self.current_values = X[self.initial_subset].sum(axis=0).astype('float64')
		else:
			raise ValueError("The initial subset must be either a two dimensional" \
				" matrix of examples or a one dimensional mask.")

		self.current_concave_values = self.concave_func(self.current_values)
		self.current_concave_values_sum = self.current_concave_values.sum()

		calculate_gains_ = calculate_gains_sparse if self.sparse else calculate_gains
		dtypes_ = sparse_dtypes if self.sparse else dtypes

		if self.optimizer in (LazyGreedy, ApproximateLazyGreedy):
			self.calculate_gains_ = calculate_gains_(self.concave_func, 
				dtypes_, False, True, False)
		elif self.optimizer in ('lazy', 'approimate-lazy'):
			self.calculate_gains_ = calculate_gains_(self.concave_func, 
				dtypes_, False, True, False)
		else: 
			self.calculate_gains_ = calculate_gains_(self.concave_func, 
				dtypes_, True, True, False)

		calculate_sieve_gains_ = calculate_gains_sieve_sparse if self.sparse else calculate_gains_sieve
		dtypes_ = sieve_sparse_dtypes if self.sparse else sieve_dtypes 
		self.calculate_sieve_gains_ = calculate_sieve_gains_(self.concave_func,
			dtypes_, True, True, False)

	def _calculate_gains(self, X, idxs=None):
		"""This function will return the gain that each example would give.

		This function will return the gains that each example would give if
		added to the selected set. When a matrix of examples is given, a
		vector will be returned showing the gain for each example. When
		a single element is passed in, it will return a singe value."""

		idxs = idxs if idxs is not None else self.idxs
		gains = numpy.zeros(idxs.shape[0], dtype='float64')

		if self.sparse:
			self.calculate_gains_(X.data, X.indices, X.indptr, gains, 
				self.current_values, self.current_concave_values, idxs)
		else:
			self.calculate_gains_(X, gains, self.current_values, idxs)
			gains -= self.current_concave_values_sum

		return gains

	def _calculate_sieve_gains(self, X, thresholds, idxs):
		"""This function will update the internal statistics from a stream.

		This function will update the various internal statistics that are a
		part of the sieve algorithm for streaming submodular optimization. This
		function does not directly return gains but it updates the values
		used by a streaming optimizer.
		"""

		super(FeatureBasedSelection, self)._calculate_sieve_gains(X,
			thresholds, idxs)

		if self.sparse:
			self.calculate_sieve_gains_(X.data, X.indices, X.indptr, 
				self.n_samples, self.sieve_current_values_, 
				self.sieve_selections_, self.sieve_gains_, 
				self.sieve_total_gains_, thresholds, 
				self.sieve_n_selected_, idxs)
		else:
			self.calculate_sieve_gains_(X, self.n_samples, 
				self.sieve_current_values_, self.sieve_selections_, 
				self.sieve_gains_, self.sieve_total_gains_, thresholds, 
				self.sieve_n_selected_, idxs)

	def _select_next(self, X, gain, idx):
		"""This function will add the given item to the selected set."""

		if self.sparse:
			self.current_values += X.toarray()[0]
		else:
			self.current_values += X

		self.current_concave_values = self.concave_func(self.current_values)
		self.current_concave_values_sum = self.current_concave_values.sum()

		super(FeatureBasedSelection, self)._select_next(
			X, gain, idx)

